'use strict';

var assert = require('assert'),
    async = require('async'),
    config = require('./config.js'),
    ejs = require('ejs'),
    EventSource = require('eventsource'),
    fs = require('fs'),
    helper = require('./helper.js'),
    https = require('https'),
    manifestFormat = require('cloudron-manifestformat'),
    opn = require('open'),
    os = require('os'),
    path = require('path'),
    ProgressBar = require('progress'),
    ProgressStream = require('progress-stream'),
    querystring = require('querystring'),
    readlineSync = require('readline-sync'),
    request = require('request'),
    safe = require('safetydance'),
    spawn = require('child_process').spawn,
    split = require('split'),
    Table = require('easy-table'),
    tar = require('tar-fs'),
    util = require('util'),
    zlib = require('zlib'),
    _ = require('underscore');

require('colors');

var exit = helper.exit;

exports = module.exports = {
    list: list,
    login: login,
    logout: logout,
    open: open,
    install: install,
    configure: configure,
    debug: debug,
    update: update,
    uninstall: uninstall,
    logs: logs,
    exec: exec,
    status: status,
    inspect: inspect,
    pull: pull,
    push: push,
    restart: restart,
    start: start,
    stop: stop,
    repair: repair,
    createOAuthAppCredentials: createOAuthAppCredentials,
    init: init,
    restore: restore,
    importApp: importApp,
    clone: clone,
    backupCreate: backupCreate,
    backupList: backupList,
    cancel: cancel,

    envGet: envGet,
    envSet: envSet,
    envList: envList,
    envUnset: envUnset
};

var NO_APP_FOUND_ERROR_STRING = 'Could not determine app. Use --app to specify the app\n';

function authenticate(adminFqdn, auth, options, callback) {
    assert.strictEqual(typeof adminFqdn, 'string');
    assert.strictEqual(typeof auth, 'object');
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    const rejectUnauthorized = !(options.parent.allowSelfsigned || options.parent.acceptSelfsigned);
    request.post(`https://${adminFqdn}/api/v1/developer/login`, { rejectUnauthorized, json: auth }, function (error, response) {
        if (error) return callback(error);
        if (response.statusCode !== 200) return callback(new Error(`Login failed: Status code: ${requestError(response)}`));

        callback(null, response.body.accessToken);
    });
}

// options for the request module
function requestOptions(options) {
    const adminFqdn = options.parent.server || config.apiEndpoint();
    const token = options.parent.token || config.token();
    const rejectUnauthorized = !(options.parent.allowSelfsigned || options.parent.acceptSelfsigned || config.allowSelfsigned());

    if (!adminFqdn && !token) return exit('Login with "cloudron login" first'); // a bit rough to put this here!

    return { adminFqdn, token, rejectUnauthorized };
}

// error for the request module
function requestError(response) {
    return `${response.statusCode} message: ${response.body.message || response.body}`; // body is sometimes just a string like in 401
}

function selectDomain(location, options, callback) {
    assert.strictEqual(typeof location, 'string');
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    request.get(`https://${adminFqdn}/api/v1/domains?access_token=${token}`, { rejectUnauthorized, json: true }, function (error, response) {
        if (error) return callback(error);
        if (response.statusCode !== 200) return callback(new Error(`Failed to list domains: ${requestError(response)}`));

        const domains = response.body.domains;

        let domain;
        let matchingDomain = domains
            .map(function (d) { return d.domain; } )
            .sort(function(a, b) { return a.length < b.length; })
            .find(function (d) { return location.endsWith(d); });

        if (matchingDomain) {
            domain = matchingDomain;
            location = location.slice(0, -matchingDomain.length-1);
        } else { // use the admin domain
            domain = domains
                .map(function (d) { return d.domain; } )
                .sort(function(a, b) { return a.length < b.length; })
                .find(function (d) { return adminFqdn.endsWith(d); });
        }

        callback(null, { location, domain });
    });
}

function stopActiveTask(app, options, callback) {
    assert.strictEqual(typeof app, 'object');
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    if (!app.taskId) return callback();

    console.log(`Stopping app's current active task ${app.taskId}`);

    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    request.post(`https://${adminFqdn}/api/v1/tasks/${app.taskId}/stop?access_token=${token}`, { json: {}, rejectUnauthorized }, function (error, response) {
        if (error) return callback(error);
        if (response.statusCode !== 204) return exit(`Failed to stop active task: ${requestError(response)}`);

        callback();
    });
}

function selectAppWithRepository(repository, options, callback) {
    assert.strictEqual(typeof repository, 'string');
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    request.get(`https://${adminFqdn}/api/v1/apps?access_token=${token}`, { rejectUnauthorized, json: true }, function (error, response) {
        if (error) return callback(error);
        if (response.statusCode !== 200) return callback(new Error(`Failed to install app: ${requestError(response)}`));

        let matchingApps = response.body.apps.filter(function (app) {
            return !app.appStoreId && app.manifest.dockerImage.startsWith(repository); // never select apps from the store
        });

        if (matchingApps.length === 0) return callback(null, [ ]);
        if (matchingApps.length === 1) return callback(null, matchingApps[0]);

        console.log();
        console.log('Available apps using same repository %s:', repository);
        matchingApps.sort(function (a, b) { return a.fqdn < b.fqdn ? -1 : 1; });
        matchingApps.forEach(function (app, index) {
            console.log('[%s]\t%s', index, app.fqdn);
        });

        var index = -1;
        // eslint-disable-next-line no-constant-condition
        while (true) {
            index = parseInt(readlineSync.question('Choose app [0-' + (matchingApps.length-1) + ']: ', {}), 10);
            if (isNaN(index) || index < 0 || index > matchingApps.length-1) console.log('Invalid selection'.red);
            else break;
        }

        callback(null, matchingApps[index]);
    });
}

// appId may be the appId or the location
function getApp(options, callback) {
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    const app = options.app || null;
    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    if (!app) { // determine based on repository name given during 'build'
        const manifestFilePath = helper.locateManifest();

        if (!manifestFilePath) return callback(new Error(NO_APP_FOUND_ERROR_STRING));

        const sourceDir = path.dirname(manifestFilePath);
        const appConfig = config.getAppConfig(sourceDir);

        if (!appConfig.repository) return callback(new Error(NO_APP_FOUND_ERROR_STRING));

        selectAppWithRepository(appConfig.repository, options, function (error, result) {
            if (error || result.length === 0) return callback(null, null);

            callback(null, result);
        });
    } else if (app.match(/.{8}-.{4}-.{4}-.{4}-.{8}/)) {  // it is an id
        request.get(`https://${adminFqdn}/api/v1/apps/${app}?access_token=${token}`, { rejectUnauthorized, json: true }, function (error, response) {
            if (error) return callback(error);
            if (response.statusCode !== 200) return callback(new Error(`Failed to get app: ${requestError(response)}`));

            callback(null, response.body);
        });
    } else { // it is a location
        request.get(`https://${adminFqdn}/api/v1/apps?access_token=${token}`, { rejectUnauthorized, json: true }, function (error, response) {
            if (error) return callback(error);
            if (response.statusCode !== 200) return callback(new Error(`Failed to get apps: ${requestError(response)}`));

            let match = response.body.apps.filter(function (m) { return m.location === app || m.fqdn === app; });
            if (match.length == 0) return callback(new Error(`App at location ${app} not found`));

            callback(null, match[0]);
        });
    }
}

function stopApp(app, options, callback) {
    assert.strictEqual(typeof app, 'object');
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/stop?access_token=${token}`, { json: {}, rejectUnauthorized }, function (error, response) {
        if (error) return callback(error);
        if (response.statusCode !== 202) return callback(`Failed to stop app: ${requestError(response)}`);

        waitForTask(response.body.taskId, options, callback);
    });
}

function startApp(app, options, callback) {
    assert.strictEqual(typeof app, 'object');
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/start?access_token=${token}`, { json: {}, rejectUnauthorized }, function (error, response) {
        if (error) return callback(error);
        if (response.statusCode !== 202) return callback(`Failed to stop app: ${requestError(response)}`);

        waitForTask(response.body.taskId, options, callback);
    });
}

function login(adminFqdn, options) {
    if (!adminFqdn) adminFqdn = readlineSync.question('Cloudron Domain (e.g. my.example.com): ', {});
    if (!adminFqdn) return exit('');

    if (adminFqdn.indexOf('https://') === 0) adminFqdn = adminFqdn.slice('https://'.length);
    if (adminFqdn.indexOf('/') !== -1) adminFqdn = adminFqdn.slice(0, adminFqdn.indexOf('/'));

    function checkIfAlreadyLoggedIn(callback) {
        config.setActive(adminFqdn);

        const token = config.token();
        if (!token) return callback('no token');

        request.get(`https://${adminFqdn}/api/v1/profile?access_token=${token}`, {}, function (error, response) {
            if (error) return callback(error);
            if (response.statusCode !== 200) return callback(`Failed to get profile: ${requestError(response)}`);

            console.log('Existing token still valid.'.green);

            callback(null, token);
        });
    }

    function done(token) {
        config.setActive(adminFqdn);
        config.setApiEndpoint(adminFqdn);
        config.setToken(token);
        config.setAllowSelfsigned(options.parent.allowSelfsigned || options.parent.acceptSelfsigned);
        config.set('cloudrons.default', adminFqdn);
    }

    checkIfAlreadyLoggedIn(function (error, token) {
        if (!error) return done(token);

        let username = readlineSync.question('Username: ', {});
        let password = readlineSync.question('Password: ', { noEchoBack: true });
        let totpToken = readlineSync.question('2FA Token (if enabled): ', {});

        authenticate(adminFqdn, { username, password, totpToken }, options, function (error, token) {
            if (error) exit(error);

            console.log('Login successful.'.green);

            done(token);
        });
    });
}

function logout() {
    config.setToken(null);
    config.set('cloudrons.default', null);

    console.log('Logged out.');
}

function open(options) {
    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        opn(`https://${app.fqdn}`);
    });
}

function list(options) {
    helper.verifyArguments(arguments);

    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    request.get(`https://${adminFqdn}/api/v1/apps?access_token=${token}`, { rejectUnauthorized, json: true }, function (error, response) {
        if (error) return exit(error);
        if (response.statusCode !== 200) return exit(`Failed to list apps: ${requestError(response)}`);

        if (response.body.apps.length === 0) return console.log('No apps installed.');

        var t = new Table();

        response.body.apps.forEach(function (app) {
            t.cell('Id', app.id);
            t.cell('Location', app.fqdn);
            t.cell('Manifest Id', (app.manifest.id || 'customapp') + '@' + app.manifest.version);
            var prettyState;
            if (app.installationState === 'installed') {
                prettyState = (app.debugMode ? 'debug' : app.runState);
            } else if (app.installationState === 'error') {
                prettyState = `error (${app.error.installationState})`;
            } else {
                prettyState = app.installationState;
            }
            t.cell('State', prettyState);
            t.newRow();
        });

        console.log();
        console.log(t.toString());
    });
}

function waitForHealthy(appId, options, callback) {
    assert.strictEqual(typeof appId, 'string');
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    if (!options.wait) return callback();

    process.stdout.write('\n => ' + 'Wait for health check '.cyan);

    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    function checkStatus() {
        request.get(`https://${adminFqdn}/api/v1/apps/${appId}?access_token=${token}`, { json: true, rejectUnauthorized }, function (error, response) {
            if (error) return callback(error);
            if (response.statusCode !== 200) return callback(new Error(`Failed to get app: ${requestError(response)}`));

            // do not check installation state here. it can be pending_backup etc (this is a bug in box code)
            if (response.body.health === 'healthy') return callback();

            // skip health check if app is in debug mode
            if (response.body.debugMode !== null) {
                process.stdout.write(' (skipped in debug mode)');
                return callback();
            }

            process.stdout.write('.');

            return setTimeout(checkStatus, 1000);
        });
    }

    setTimeout(checkStatus, 1000);
}

function waitForTask(taskId, options, callback) {
    assert.strictEqual(typeof taskId, 'string');
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    let currentMessage = '';

    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    function checkStatus() {
        request.get(`https://${adminFqdn}/api/v1/tasks/${taskId}?access_token=${token}`, { json: true, rejectUnauthorized }, function (error, response) {
            if (error) return callback(error);
            if (response.statusCode !== 200) return callback(new Error(`Failed to get task: ${requestError(response)}`));

            // note: for queued tasks, 'active' returns false
            if (response.body.error || response.body.percent === 100) return callback(null, response.body); // task errored or done

            let message = response.body.message || '';

            // Backup downloading includes the filename and clutters the cli output
            if (message.indexOf('Restore - Down') === 0) {
                message = 'Downloading backup';
            }

            // track current progress and show progress dots
            if (currentMessage && currentMessage === message) {
                if (currentMessage.indexOf('Creating image') === -1 && currentMessage.indexOf('Downloading backup') === -1) process.stdout.write('.');
            } else if (message) {
                process.stdout.write('\n => ' + message.trim().cyan + ' ');
            } else {
                process.stdout.write('\n => ' + 'Waiting to start installation '.cyan);
            }

            currentMessage = message;

            setTimeout(checkStatus, 1000);
        });
    }

    checkStatus();
}

function waitForFinishInstallation(appId, taskId, options, callback) {
    assert.strictEqual(typeof appId, 'string');
    assert.strictEqual(typeof taskId, 'string');
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    waitForTask(taskId, options, function (error) {
        if (error) return callback(error);

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

        request.get(`https://${adminFqdn}/api/v1/apps/${appId}?access_token=${token}`, { json: true, rejectUnauthorized }, function (error, response) {
            if (error) return callback(error);
            if (response.statusCode !== 200) return callback(new Error(`Failed to get app: ${requestError(response)}`));

            if (response.body.installationState !== 'installed') return callback(new Error(`Installation failed: ${response.body.error ? response.body.error.message : ''}`));

            waitForHealthy(appId, options, callback);
        });
    });
}

function queryPortBindings(app, manifest) {
    var portBindings = { };
    const allPorts = _.extend({}, manifest.tcpPorts, manifest.udpPorts);
    for (var env in allPorts) {
        var defaultPort = (app && app.portBindings && app.portBindings[env]) ? app.portBindings[env] : (allPorts[env].defaultValue || '');
        var port = readlineSync.question(allPorts[env].description + ' (default ' + env + '=' + defaultPort + '. "x" to disable): ', {});
        if (port === '') {
            portBindings[env] = defaultPort;
        } else if (isNaN(parseInt(port, 10))) {
            console.log(('Cleared ' + env).gray);
        } else {
            portBindings[env] = parseInt(port, 10);
        }
    }
    return portBindings;
}

// this function can fail on many levels
function parseDebugCommand(cmd) {
    // hack for commander
    if (typeof cmd !== 'string' || cmd === '') return [ '/bin/bash', '-c', 'echo "Debug mode. Use cloudron exec to debug. Sleeping" && sleep infinity' ];

    if (cmd == 'default') return null; // another hack

    return cmd.split(' '); // yet another hack
}

function downloadManifest(appstoreId, callback) {
    const parts = appstoreId.split('@');
    const url = config.appStoreOrigin() + '/api/v1/apps/' + parts[0] + (parts[1] ? '/versions/' + parts[1] : '');

    request.get(url, { json: true }, function (error, response) {
        if (error) return callback(error);
        if (response.statusCode !== 200) return exit(`Failed to get app info from store: ${requestError(response)}`);

        callback(null, response.body.manifest, null /* manifest file path */);
    });
}

function getManifest(appstoreId, callback) {
    assert.strictEqual(typeof appstoreId, 'string');
    assert.strictEqual(typeof callback, 'function');

    if (appstoreId) return downloadManifest(appstoreId, callback);

    var manifestFilePath = helper.locateManifest();
    if (!manifestFilePath) return callback(new Error('No CloudronManifest.json found'));

    var result = manifestFormat.parseFile(manifestFilePath);
    if (result.error) return exit('Invalid CloudronManifest.json: '.red + result.error.message);

    // resolve post install message
    if (result.manifest.postInstallMessage && result.manifest.postInstallMessage.slice(0, 7) === 'file://') {
        var postInstallFilename = result.manifest.postInstallMessage.slice(7);
        // resolve filename wrt manifest
        if (manifestFilePath) postInstallFilename = path.resolve(path.dirname(manifestFilePath), postInstallFilename);
        result.manifest.postInstallMessage = safe.fs.readFileSync(postInstallFilename, 'utf8');
        if (!result.manifest.postInstallMessage) return exit('Could not read postInstallMessage file ' + postInstallFilename + ':' + safe.error.message);
    }

    callback(null, result.manifest, manifestFilePath);
}

function install(options) {
    helper.verifyArguments(arguments);

    if (options.app) exit('Use "cloudron update" to update an app');

    getManifest(options.appstoreId || '', function (error, manifest, manifestFilePath) {
        if (error) return exit(error);

        if (!manifest.dockerImage) {
            const sourceDir = path.dirname(manifestFilePath);
            const image = options.image || config.getAppConfig(sourceDir).dockerImage;

            if (!image) exit('No image found, please run `cloudron build` first or specify using --image');

            manifest.dockerImage = image;
        }

        const location = options.location || readlineSync.question('Location: ', {});
        if (!location) return exit('');

        selectDomain(location, options, function (error, domainObject) {
            if (error) return exit(error);

            // port bindings
            let portBindings = {};
            if (options.portBindings) {
                // ask the user for port values if the ports are different in the app and the manifest
                if (typeof options.portBindings === 'string') {
                    portBindings = { };
                    options.portBindings.split(',').forEach(function (kv) {
                        var tmp = kv.split('=');
                        if (isNaN(parseInt(tmp[1], 10))) return; // disable the port
                        portBindings[tmp[0]] = parseInt(tmp[1], 10);
                    });
                } else {
                    portBindings = queryPortBindings(null /* existing app */, manifest);
                }
            } else { // just put in defaults
                const allPorts = _.extend({}, manifest.tcpPorts, manifest.udpPorts);
                for (let portName in allPorts) {
                    portBindings[portName] = allPorts[portName].defaultValue;
                }
            }

            for (var binding in portBindings) {
                console.log('%s: %s', binding, portBindings[binding]);
            }

            let data = {
                appStoreId: options.appstoreId || '', // note case change
                manifest: options.appstoreId ? null : manifest, // cloudron ignores manifest anyway if appStoreId is set
                location: domainObject.location,
                domain: domainObject.domain,
                portBindings: portBindings,
                accessRestriction: null
            };

            // the sso only applies for apps which allow optional sso
            if (manifest.optionalSso) data.sso = options.sso;

            if (options.debug) {
                data.debugMode = {
                    readonlyRootfs: options.readonly ? true : false,
                    cmd: parseDebugCommand(options.debug)
                };
                data.memoryLimit = -1;
                options.wait = false; // in debug mode, health check never succeeds
            }

            if (!options.appstoreId && manifest.icon) {
                let iconFilename = manifest.icon.slice(0, 7) === 'file://' ? manifest.icon.slice(7) : manifest.icon;
                iconFilename = path.resolve(path.dirname(manifestFilePath), iconFilename); // resolve filename wrt manifest
                data.icon = safe.fs.readFileSync(iconFilename, { encoding: 'base64' });
                if (!data.icon) return exit(`Could not read icon: ${iconFilename} message: ${safe.error.message}`);
            }

            const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

            request.post(`https://${adminFqdn}/api/v1/apps/install?access_token=${token}`, { rejectUnauthorized, json: data }, function (error, response) {
                if (error) return exit(error);
                if (response.statusCode !== 202) return exit(`Failed to install app: ${requestError(response)}`);

                const appId = response.body.id;

                console.log('App is being installed.');

                waitForFinishInstallation(appId, response.body.taskId, options, function (error) {
                    if (error) exit('\n\nApp installation error: %s'.red, error.message);

                    console.log('\n\nApp is installed.'.green);
                });
            });
        });
    });
}

function configure(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        if (!options.location) options.location = readlineSync.question(`Enter new location (default: ${app.fqdn}): `, { });
        let location = options.location || app.location;

        selectDomain(location, options, function (error, domainObject) {
            if (error) exit(error);

            const data = {
                location: domainObject.location,
                domain: domainObject.domain,
                portBindings: app.portBindings
            };

            // port bindings
            if (options.portBindings) {
                var portBindings = app.portBindings;
                // ask the user for port values if the ports are different in the app and the manifest
                if (typeof options.portBindings === 'string') {
                    portBindings = {};
                    options.portBindings.split(',').forEach(function (kv) {
                        var tmp = kv.split('=');
                        if (isNaN(parseInt(tmp[1], 10))) return; // disable the port
                        portBindings[tmp[0]] = parseInt(tmp[1], 10);
                    });
                } else {
                    portBindings = queryPortBindings(app, app.manifest);
                }

                for (let binding in portBindings) {
                    console.log('%s: %s', binding, portBindings[binding]);
                }

                data.portBindings = portBindings;
            }

            const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

            request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/configure/location?access_token=${token}`, { rejectUnauthorized, json: data }, function (error, response) {
                if (error) return exit(error);
                if (response.statusCode !== 202) return exit(`Failed to configure app: ${requestError(response)}`);

                waitForTask(response.body.taskId, options, function (error) {
                    if (error) return exit(error);

                    waitForHealthy(app.id, options, function (error) {
                        if (error) return exit(error);

                        console.log('\n\nApp configured'.green);
                    });
                });
            });
        });
    });
}

function update(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        getManifest('' /* appStoreId */, function (error, manifest, manifestFilePath) {
            if (error) return exit(error);

            const sourceDir = path.dirname(manifestFilePath);
            const image = options.image || config.getAppConfig(sourceDir).dockerImage;

            if (!image) return exit('No image found, please run `cloudron build` first or use --image');

            manifest.dockerImage = image;

            const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);
            let url, data;

            if (app.error && (app.error.installationState === 'pending_install')) { // install had failed. call repair to re-install
                url = `https://${adminFqdn}/api/v1/apps/${app.id}/repair?access_token=${token}`;
                data = { manifest };
            } else {
                url = `https://${adminFqdn}/api/v1/apps/${app.id}/update?access_token=${token}`;
                data = {
                    manifest: manifest,
                    skipBackup: !options.backup,
                    skipNotification: true,
                    force: true // permit downgrades
                };
            }

            if (manifest.icon) {
                let iconFilename = manifest.icon.slice(0, 7) === 'file://' ? manifest.icon.slice(7) : manifest.icon;
                iconFilename = path.resolve(path.dirname(manifestFilePath), iconFilename); // resolve filename wrt manifest
                data.icon = safe.fs.readFileSync(iconFilename, { encoding: 'base64' });
                if (!data.icon) return exit(`Could not read icon: ${iconFilename} message: ${safe.error.message}`);
            }

            request.post(url, { rejectUnauthorized, json: data }, function (error, response) {
                if (error) return exit(error);
                if (response.statusCode !== 202) return exit(`Failed to update app: ${requestError(response)}`);

                process.stdout.write('\n => ' + 'Waiting for app to be updated '.cyan);

                waitForFinishInstallation(app.id, response.body.taskId, options, function (error) {
                    if (error) exit('\n\nApp update error: %s'.red, error.message);

                    console.log('\n\nApp is updated.'.green);
                });
            });
        });
    });
}

function debug(cmd, options) {
    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        var data = {
            debugMode: options.disable ? null : {
                readonlyRootfs: options.readonly ? true : false,
                cmd: parseDebugCommand(cmd.join(' ').trim())
            }
        };

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

        request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/configure/debug_mode?access_token=${token}`, { rejectUnauthorized, json: data }, function (error, response) {
            if (error) return exit(error);
            if (response.statusCode !== 202) return exit(`Failed to set debug mode: ${requestError(response)}`);

            waitForTask(response.body.taskId, options, function (error) {
                if (error) return exit(error);

                var memoryLimit = options.limitMemory ? 0 : -1;

                // skip setting memory limit if unchanged
                if (app.memoryLimit === memoryLimit) return console.log('\n\nDone');

                console.log('\n');
                console.log(options.limitMemory ? 'Limiting memory' : 'Setting unlimited memory');

                request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/configure/memory_limit?access_token=${token}`, { rejectUnauthorized, json: { memoryLimit } }, function (error, response) {
                    if (error) return exit(error);
                    if (response.statusCode !== 202) return exit(`Failed to set memory limit: ${requestError(response)}`);

                    waitForTask(response.body.taskId, options, function (error) {
                        if (error) return exit(error);

                        console.log('\n\nDone');
                    });
                });
            });
        });
    });
}

function repair(options) {
    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        getManifest('' /* appStoreId */, function (error, manifest, manifestFilePath) {
            let dockerImage = options.image;
            if (!dockerImage) {
                if (error) return exit(error);

                const sourceDir = path.dirname(manifestFilePath);
                dockerImage = config.getAppConfig(sourceDir).dockerImage;
            }

            if (!dockerImage) return exit('No image found, please run `cloudron build` first or use --image');

            const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);
            const data = { dockerImage };

            request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/repair?access_token=${token}`, { rejectUnauthorized, json: data }, function (error, response) {
                if (error) return exit(error);
                if (response.statusCode !== 202) return exit(`Failed to set repair mode: ${requestError(response)}`);

                process.stdout.write('\n => ' + 'Waiting for app to be repaired '.cyan);

                waitForFinishInstallation(app.id, response.body.taskId, options, function (error) {
                    if (error) exit('\n\nApp repair error: %s'.red, error.message);

                    console.log('\n\nApp is repaired.'.green);
                });
            });
        });
    });
}

function cancel(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        if (!app.taskId) return exit('No active task.');

        stopActiveTask(app, options, function (error) {
            if (error) exit(error);

            console.log('\nTask stopped.'.green);
        });
    });
}

function uninstall(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        stopActiveTask(app, options, function (error) {
            if (error) exit(error);

            const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

            request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/uninstall?access_token=${token}`, { rejectUnauthorized, json: true }, function (error, response) {
                if (error) return exit(error);
                if (response.statusCode !== 202) return exit(`Failed to uninstall app: ${requestError(response)}`);

                process.stdout.write('\n => ' + 'Waiting for app to be uninstalled '.cyan);

                waitForTask(response.body.taskId, options, function (error) {
                    if (error) return exit(error);

                    request.get(`https://${adminFqdn}/api/v1/apps/${app.id}?access_token=${token}`, { rejectUnauthorized, json: true }, function (error, response) {
                        if (error) return exit(error);
                        if (response.statusCode === 404) {
                            console.log('\n\nApp %s successfully uninstalled.', app.fqdn.bold);
                        } else if (response.body.installationState === 'error') {
                            console.log('\n\nApp uninstallation failed.\n'.red);
                            exit(response.body.errorMessage);
                        }
                    });
                });
            });
        });
    });
}

function logPrinter(obj) {
    let message;

    if (obj.message === null) {
        message = '[large binary blob skipped]';
    } else if (typeof obj.message === 'string') {
        message = obj.message;
    } else if (util.isArray(obj.message)) {
        message = Buffer.from(obj.message).toString('utf8');
    }

    var ts = new Date(obj.realtimeTimestamp/1000).toTimeString().split(' ')[0];
    console.log('%s - %s', ts.cyan, message);
}

function logs(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);
        const lines = options.lines || 500;

        if (!options.tail) {
            const req = request.get(`https://${adminFqdn}/api/v1/apps/${app.id}/logs?access_token=${token}&lines=${lines}`, { rejectUnauthorized }, function (error, response) {
                if (error) return exit(error);

                if (response.statusCode !== 200) return exit(`Failed to get logs: ${requestError(response)}`);
            });

            req.pipe(split(JSON.parse))
                .on('data', logPrinter)
                .on('error', process.exit)
                .on('end', process.exit);

            return;
        }

        var es = new EventSource(`https://${adminFqdn}/api/v1/apps/${app.id}/logstream?access_token=${token}&lines=10`, { rejectUnauthorized }); // not sure why this is needed

        es.on('message', function (e) { // e { type, data, lastEventId }. lastEventId is the timestamp
            logPrinter(JSON.parse(e.data));
        });

        es.on('error', function (error) {
            if (error.status === 401) return authenticate({ error: true }, logs.bind(null, options));
            if (error.status === 412) exit('Logs currently not available. App is not installed.');
            exit(error);
        });
    });
}

function status(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        console.log();
        console.log('Id: \t\t', app.id.cyan);
        console.log('Location: \t', app.fqdn.cyan);
        console.log('Version: \t', app.manifest.version.cyan);
        console.log('Manifest Id: \t', (app.manifest.id || 'customapp').cyan);
        console.log('Install state: \t', (app.installationState + (app.debugMode ? ' (debug)' : '')).cyan);
        console.log('Run state: \t', app.runState.cyan);
        console.log();

        exit();
    });
}

function inspect(options) {
    helper.verifyArguments(arguments);

    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    request.get(`https://${adminFqdn}/api/v1/apps?access_token=${token}`, { rejectUnauthorized, json: true }, function (error, response) {
        if (error) return exit(error);
        if (response.statusCode !== 200) return exit(`Failed to list apps: ${requestError(response)}`);

        let apps = [];

        async.eachSeries(response.body.apps, function (app, iteratorDone) {
            request.get(`https://${adminFqdn}/api/v1/apps/${app.id}?access_token=${token}`, { rejectUnauthorized, json: true }, function (error, response) {
                if (error) return iteratorDone(error);
                if (response.statusCode !== 200) return iteratorDone(`Failed to list app: ${requestError(response)}`);

                apps.push(response.body);
                iteratorDone();
            });
        }, function (error) {
            if (error) exit(error);

            console.log(JSON.stringify({
                apiEndpoint: adminFqdn,
                appStoreOrigin: config.appStoreOrigin(),
                apps: apps
            }, null, 4));
        });
    });
}

function restart(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        stopApp(app, options, function (error) {
            if (error) exit(error);

            startApp(app, options, function (error) {
                if (error) return exit(error);

                waitForHealthy(app.id, options, function (error) {
                    if (error) return exit(error);

                    console.log('\n\nApp restarted'.green);
                });
            });
        });
    });
}

function start(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        startApp(app, options, function (error) {
            if (error) return exit(error);

            waitForHealthy(app.id, options, function (error) {
                if (error) return exit(error);

                console.log('\n\nApp started'.green);
            });
        });
    });
}

function stop(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        stopApp(app, options, function (error) {
            if (error) exit(error);

            console.log('\n\nApp stopped'.green);

            exit();
        });
    });
}

function backupCreate(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

        // get existing options and then remove
        request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/backup?access_token=${token}`, { json: {}, rejectUnauthorized }, function (error, response) {
            if (error) return exit(error);
            if (response.statusCode !== 202) return exit(`Failed to start backup: ${requestError(response)}`);

            // FIXME: this should be waitForHealthCheck but the box code incorrectly modifies the installationState
            waitForFinishInstallation(app.id, response.body.taskId, options, function (error) {
                if (error) return exit('\n\nApp backup error: %s'.red, error.message);

                console.log('\n\nApp is backed up'.green);
            });
        });
    });
}

function backupList(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

        // get existing options and then remove
        request.get(`https://${adminFqdn}/api/v1/apps/${app.id}/backups?access_token=${token}`, { json: true, rejectUnauthorized }, function (error, response) {
            if (error) return exit(error);
            if (response.statusCode !== 200) return exit(`Failed to list backups: ${requestError(response)}`);

            if (options.raw) return console.log(JSON.stringify(response.body.backups, null, 4));

            if (response.body.backups.length === 0) return console.log('No backups.');

            response.body.backups = response.body.backups.map(function (backup) {
                backup.creationTime = new Date(backup.creationTime);
                return backup;
            }).sort(function (a, b) { return b.creationTime - a.creationTime; });

            var t = new Table();

            response.body.backups.forEach(function (backup) {
                t.cell('Id', backup.id);
                t.cell('Creation Time', backup.creationTime);
                t.cell('Version', backup.version);

                t.newRow();
            });

            console.log();
            console.log(t.toString());
        });
    });
}

function getLastBackup(app, options, callback) {
    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    // get existing options and then remove
    request.get(`https://${adminFqdn}/api/v1/apps/${app.id}/backups?access_token=${token}`, { json: true, rejectUnauthorized }, function (error, response) {
        if (error) return exit(error);
        if (response.statusCode !== 200) return callback(new Error(`Failed to list backups: ${requestError(response)}`));
        if (response.body.backups.length === 0) return callback(new Error('No backups'));

        response.body.backups = response.body.backups.map(function (backup) {
            backup.creationTime = new Date(backup.creationTime);
            return backup;
        }).sort(function (a, b) { return b.creationTime - a.creationTime; });

        return callback(null, response.body.backups[0].id);
    });
}

function restore(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        var func = !options.backup ? getLastBackup.bind(null, app, options) : function (next) { next(null, options.backup); };

        func(function (error, backupId) {
            if (error) return exit(error);

            const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

            // get existing options and then remove
            request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/restore?access_token=${token}`, { json: { backupId }, rejectUnauthorized }, function (error, response) {
                if (error) return exit(error);
                if (response.statusCode !== 202) return exit(`Failed to restore app: ${requestError(response)}`);

                // FIXME: this should be waitForHealthCheck but the box code incorrectly modifies the installationState
                waitForFinishInstallation(app.id, response.body.taskId, options, function (error) {
                    if (error) {
                        return exit('\n\nApp restore error: %s'.red, error.message);
                    }

                    console.log('\n\nApp is restored'.green);
                });
            });
        });
    });
}

function importApp(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);
        let data = {};

        if (!options.inPlace) {
            let { backupId, backupFormat, backupConfig, backupPath, backupKey } = options;
            let backupFolder;

            if (backupPath) {
                if (backupPath.endsWith('.tar.gz')) {
                    backupFolder = path.dirname(backupPath);
                    backupId = path.basename(backupPath).replace(/.tar.gz$/, '');
                    backupFormat = 'tgz';
                } else {
                    backupFolder = path.dirname(backupPath);
                    backupId = path.basename(backupPath);
                    backupFormat = 'rsync';
                }
                backupConfig = { provider: 'filesystem', backupFolder };
                data = { backupId, backupFormat, backupConfig };
            } else {
                if (!backupId) return exit('--backup-id, --backup-file or --in-place is required');
                if (!backupFormat) return exit('--backup-format is required');

                if (backupId.endsWith('.tar.gz')) { // common mistake...
                    return exit('backupId should not end with .tar.gz. Try again removing the extension from backup id');
                }

                if (backupConfig) {
                    backupConfig = safe.JSON.parse(options.backupConfig);
                    if (!backupConfig) return exit('--backup-config is not a valid JSON');
                } else {
                    backupConfig = null; // ensure null-ness
                }

                data = { backupId, backupFormat, backupConfig };
            }

            if (backupKey) data.backupConfig.key = backupKey;
        }

        request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/import?access_token=${token}`, { rejectUnauthorized, json: data }, function (error, response) {
            if (error) return exit(error);
            if (response.statusCode !== 202) return exit(`Failed to import app: ${requestError(response)}`);

            waitForFinishInstallation(app.id, response.body.taskId, options, function (error) {
                if (error) return exit('\n\nApp import error: %s'.red, error.message);

                console.log('\n\nApp is restored'.green);
            });
        });
    });
}

function clone(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);
        if (!options.backup) return exit('Use --backup to specify the backup id');

        let location = options.location || readlineSync.question('Cloned app location: ', {});
        let portBindings = queryPortBindings(app, app.manifest);
        let backupId = options.backup;

        selectDomain(location, options, function (error, domainObject) {
            if (error) return exit(error);

            const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

            const data = { backupId, domain: domainObject.domain, location: domainObject.location, portBindings: portBindings };

            request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/clone?access_token=${token}`, { rejectUnauthorized, json: data }, function (error, response) {
                if (error) return exit(error);
                if (response.statusCode !== 201) return exit(`Failed to list apps: ${requestError(response)}`);

                // FIXME: this should be waitForHealthCheck but the box code incorrectly modifies the installationState
                console.log('App cloned as id ' + response.body.id);
                waitForFinishInstallation(response.body.id, response.body.taskId, options, function (error) {
                    if (error) return exit(`\n\nClone error: ${error.message}`);

                    console.log('\n\nApp is cloned'.green);
                });
            });
        });
    });
}

// taken from docker-modem
function demuxStream(stream, stdout, stderr) {
    var header = null;

    stream.on('readable', function() {
        header = header || stream.read(8);
        while (header !== null) {
            var type = header.readUInt8(0);
            var payload = stream.read(header.readUInt32BE(4));
            if (payload === null) break;
            if (type == 2) {
                stderr.write(payload);
            } else {
                stdout.write(payload);
            }
            header = stream.read(8);
        }
    });
}

// cloudron exec - must work interactively. needs tty.
// cloudron exec -- ls asdf  - must work
// cloudron exec -- cat /home/cloudron/start.sh > /tmp/start.sh - must work (test with binary files). should disable tty
// echo "sauce" | cloudron exec -- bash -c "cat - > /app/data/sauce" - test with binary files. should disable tty
// cat ~/tmp/fantome.tar.gz | cloudron exec -- bash -c "tar xvf - -C /tmp" - must show an error
// cat ~/tmp/fantome.tar.gz | cloudron exec -- bash -c "tar zxf - -C /tmp" - must extrack ok
function exec(cmd, options) {
    var stdin = options._stdin || process.stdin; // hack for 'push', 'pull' to reuse this function
    var stdout = options._stdout || process.stdout;

    var tty = !!options.tty;

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        if (app.installationState !== 'installed') exit('App is not yet running. Try again later.');

        if (cmd.length === 0) {
            cmd = [ '/bin/bash' ];
            tty = true; // override
        }

        if (tty && !stdin.isTTY) exit('stdin is not tty');

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

        var query = {
            rows: stdout.rows,
            columns: stdout.columns,
            access_token: token,
            cmd: JSON.stringify(cmd),
            tty: tty
        };

        var req = https.request({
            hostname: adminFqdn,
            path: `/api/v1/apps/${app.id}/exec?` + querystring.stringify(query),
            method: 'GET',
            headers: {
                'Connection': 'Upgrade',
                'Upgrade': 'tcp'
            },
            rejectUnauthorized
        }, function handler(res) {
            if (res.statusCode === 403) exit('Unauthorized.'.red); // only admin or only owner (for docker addon)

            exit('Could not upgrade connection to tcp. http status:', res.statusCode);
        });

        req.on('upgrade', function (resThatShouldNotBeUsed, socket /*, upgradeHead*/) {
            // do not use res here! it's all socket from here on
            socket.on('error', exit);

            socket.setNoDelay(true);
            socket.setKeepAlive(true);

            if (tty) {
                stdin.setRawMode(true);
                stdin.pipe(socket, { end: false }); // the remote will close the connection
                socket.pipe(stdout); // in tty mode, stdout/stderr is merged
                socket.on('end', exit); // server closed the socket
            } else {// create stdin process on demand
                if (typeof stdin === 'function') stdin = stdin();

                stdin.on('data', function (d) {
                    var buf = Buffer.alloc(4);
                    buf.writeUInt32BE(d.length, 0 /* offset */);
                    socket.write(buf);
                    socket.write(d);
                });
                stdin.on('end', function () {
                    var buf = Buffer.alloc(4);
                    buf.writeUInt32BE(0, 0 /* offset */);
                    socket.write(buf);
                });

                demuxStream(socket, stdout, process.stderr); // can get separate streams in non-tty mode
                socket.on('end', function () {  // server closed the socket
                    stdin.end(); // required for this process to 'exit' cleanly. do not call exit() because writes may not have finished
                    if (stdout !== process.stdout) stdout.end(); // for push stream

                    socket.end();
                    // process._getActiveHandles(); process._getActiveRequests();
                    setImmediate(() => process.exit()); // for some reason the write socket doesn't close in latest node
                });
            }
        });

        req.on('error', exit); // could not make a request
        req.end(); // this makes the request
    });
}

function push(localDir, remote, options) {
    // deal with paths prefixed with ~
    var local = localDir.replace(/^~(?=$|\/|\\)/, os.homedir());
    var stat = fs.existsSync(path.resolve(local)) ? fs.lstatSync(local) : null;

    if (stat && stat.isDirectory())  {
        // Create a functor for stdin. If no data event handlers are attached, and there are no stream.pipe() destinations, and the stream is
        // switched into flowing mode, then data will be lost. So, we have to start the tarzip only when exec is ready to attach event handlers.
        options._stdin = function () {
            var tarzip = spawn('tar', ['zcf', '-', '-C', path.dirname(local), path.basename(local)], { stdio: 'pipe' });
            return tarzip.stdout;
        };

        exec(['tar', 'zxvf', '-', '-C', remote], options);
    } else {
        if (local === '-') {
            options._stdin = process.stdin;
        } else if (stat) {
            var progress = new ProgressStream({ length: stat.size, time: 1000 });

            options._stdin = progress;
            fs.createReadStream(local).pipe(progress);

            var bar = new ProgressBar('Uploading [:bar] :percent: :etas', {
                complete: '=',
                incomplete: ' ',
                width: 100,
                total: stat.size
            });

            progress.on('progress', function (p) { bar.update(p.percentage / 100); /* bar.tick(p.transferred - bar.curr); */ });
        } else {
            exit('local file ' + local + ' does not exist');
        }

        options._stdin.on('error', function (error) { exit('Error pushing', error); });

        if (remote.endsWith('/')) { // dir
            remote = remote + '/' + path.basename(local); // do not use path.join as we want this to be a UNIX path
        }

        exec(['bash', '-c', `cat - > "${remote}"`], options);
    }
}

function pull(remote, local, options) {
    if (remote.endsWith('/')) { // dir
        var untar = tar.extract(local); // local directory is created if it doesn't exist!
        var unzip = zlib.createGunzip();

        unzip.pipe(untar);
        options._stdout = unzip;

        exec(['tar', 'zcf', '-', '-C', remote, '.'], options);
    } else {
        if (fs.existsSync(local) && fs.lstatSync(local).isDirectory()) {
            local = path.join(local, path.basename(remote));
            options._stdout = fs.createWriteStream(local);
        } else if (local === '-') {
            options._stdout = process.stdout;
        } else {
            options._stdout = fs.createWriteStream(local);
        }

        options._stdout.on('error', function (error) { exit('Error pulling', error); });

        exec(['cat', remote], options);
    }
}

function createOAuthAppCredentials(options) {
    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    const data = { appId: 'localdevelopment', redirectURI: options.redirectUri, scope: options.scope };

    request.post(`https://${adminFqdn}/api/v1/clients?access_token=${token}`, { json: data, rejectUnauthorized }, function (error, response) {
        if (error) return exit(error);
        if (response.statusCode !== 201) return exit(`Failed to create oauth app credentials: ${requestError(response)}`);

        if (options.shell) {
            console.log('export CLOUDRON_CLIENT_ID="%s";\nexport CLOUDRON_CLIENT_SECRET="%s";\nexport CLOUDRON_REDIRECT_URI="%s"\nexport CLOUDRON_API_ORIGIN="%s"',
                response.body.id, response.body.clientSecret, response.body.redirectURI, `https://${adminFqdn}`);
        } else {
            console.log();
            console.log('Useful OAuth information:'.bold);
            console.log(`apiOrigin:        https://${adminFqdn}`);
            console.log(`authorizationURL: https://${adminFqdn}/api/v1/oauth/dialog/authorize`);
            console.log(`tokenURL:         https://${adminFqdn}/api/v1/oauth/token`);
            console.log();
            console.log('New oauth app credentials:'.bold);
            console.log('ClientId:         %s', response.body.id.cyan);
            console.log('ClientSecret:     %s', response.body.clientSecret.cyan);
            console.log('RedirectURI:      %s', response.body.redirectURI.cyan);
            console.log();
            console.log('Gulp development:'.bold);
            console.log(`gulp develop --api-origin https://${adminFqdn} --client-id ${response.body.id} --client-secret ${response.body.clientSecret}`);
        }
    });
}

function init() {
    var manifestFilePath = helper.locateManifest();
    if (manifestFilePath && path.dirname(manifestFilePath) === process.cwd()) return exit('CloudronManifest.json already exists in current directory'.red);

    var manifestTemplate = fs.readFileSync(path.join(__dirname, 'templates/', 'CloudronManifest.json.ejs'), 'utf8');
    var dockerfileTemplate = fs.readFileSync(path.join(__dirname, 'templates/', 'Dockerfile.ejs'), 'utf8');
    var dockerignoreTemplate = fs.readFileSync(path.join(__dirname, 'templates/', 'dockerignore.ejs'), 'utf8');

    var data = {
        version: '0.1.0',
        title: 'App title',
        httpPort: 3000
    };

    var manifest = ejs.render(manifestTemplate, data);
    fs.writeFileSync('CloudronManifest.json', manifest, 'utf8');

    if (fs.existsSync('Dockerfile')) {
        console.log('Dockerfile already exists, skipping');
    } else {
        var dockerfile = ejs.render(dockerfileTemplate, data);
        fs.writeFileSync('Dockerfile', dockerfile, 'utf8');
    }

    if (fs.existsSync('.dockerignore')) {
        console.log('.dockerignore already exists, skipping');
    } else {
        var dockerignore = ejs.render(dockerignoreTemplate, data);
        fs.writeFileSync('.dockerignore', dockerignore, 'utf8');
    }

    console.log();
    console.log('Now edit the CloudronManifest.json'.yellow.bold);
    console.log();
}

function setEnv(app, env, options, callback) {
    assert.strictEqual(typeof app, 'object');
    assert.strictEqual(typeof env, 'object');
    assert.strictEqual(typeof options, 'object');
    assert.strictEqual(typeof callback, 'function');

    const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

    // get existing options first and then merge
    request.post(`https://${adminFqdn}/api/v1/apps/${app.id}/configure/env?access_token=${token}`, { json: { env }, rejectUnauthorized }, function (error, response) {
        if (error) return exit(error);
        if (response.statusCode !== 202) return exit(`Failed to set env: ${requestError(response)}`);

        waitForTask(response.body.taskId, options, function (error) {
            if (error) return callback(error);

            console.log('\n');
        });
    });
}

function envSet(envVars, options) {
    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

        // get existing options first and then merge
        request.get(`https://${adminFqdn}/api/v1/apps/${app.id}?access_token=${token}`, { json: true, rejectUnauthorized }, function (error, response) {
            if (error) return exit(error);
            if (response.statusCode !== 200) return exit(`Failed to get app: ${requestError(response)}`);

            let env = response.body.env;
            envVars.forEach(envVar => {
                let m = envVar.match(/(.*?)=(.*)/);
                if (!m) return exit(`Expecting KEY=VALUE pattern. Got ${envVar}`);
                env[m[1]] = m[2];
            });

            setEnv(app, env, options, exit);
        });
    });
}

function envUnset(envNames, options) {
    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

        // get existing options and then remove
        request.get(`https://${adminFqdn}/api/v1/apps/${app.id}?access_token=${token}`, { json: true, rejectUnauthorized }, function (error, response) {
            if (error) return exit(error);
            if (response.statusCode !== 200) return exit(`Failed to get app: ${requestError(response)}`);

            let env = response.body.env;
            envNames.forEach(name => delete env[name]);

            setEnv(app, env, options, exit);
        });
    });
}

function envList(options) {
    helper.verifyArguments(arguments);

    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

        request.get(`https://${adminFqdn}/api/v1/apps/${app.id}?access_token=${token}`, { json: true, rejectUnauthorized }, function (error, response) {
            if (error) return exit(error);
            if (response.statusCode !== 200) return exit(`Failed to get app: ${requestError(response)}`);

            var t = new Table();

            const envNames = Object.keys(response.body.env);
            if (envNames.length === 0) return console.log('No custom environment variables');

            envNames.forEach(function (envName) {
                t.cell('Name', envName);
                t.cell('Value', response.body.env[envName]);

                t.newRow();
            });

            console.log();
            console.log(t.toString());
        });
    });
}

function envGet(envName, options) {
    getApp(options, function (error, app) {
        if (error) return exit(error);

        if (!app) return exit(NO_APP_FOUND_ERROR_STRING);

        const { adminFqdn, token, rejectUnauthorized } = requestOptions(options);

        request.get(`https://${adminFqdn}/api/v1/apps/${app.id}?access_token=${token}`, { json: true, rejectUnauthorized }, function (error, response) {
            if (error) return exit(error);
            if (response.statusCode !== 200) return exit(`Failed to get app: ${requestError(response)}`);

            console.log(response.body.env[envName]);
        });
    });
}
